"""Provide high-level UDP endpoints for asyncio.

Copied/modified from https://gist.github.com/vxgmichel/e47bff34b68adb3cf6bd4845c4bed448
"""
from __future__ import annotations

__all__ = ["open_local_endpoint", "open_remote_endpoint"]

import asyncio
import warnings
from typing import Any

# pylint: disable=protected-access

# Datagram protocol


class DatagramEndpointProtocol(asyncio.DatagramProtocol):
    """Datagram protocol for the endpoint high-level interface."""

    def __init__(self, endpoint: Endpoint) -> None:
        self._endpoint = endpoint

    # Protocol methods

    def connection_made(self, transport: Any) -> None:
        self._endpoint._transport = transport

    def connection_lost(self, exc: Any) -> None:
        assert exc is None
        if self._endpoint._write_ready_future is not None:  # pragma: no cover
            self._endpoint._write_ready_future.set_result(None)
        self._endpoint.close()

    # Datagram protocol methods

    def datagram_received(self, data: bytes, addr: tuple[str, int]) -> None:
        self._endpoint.feed_datagram(data, addr)

    def error_received(self, exc: Any) -> None:
        msg = "Endpoint received an error: {!r}"
        warnings.warn(msg.format(exc))

    # Workflow control

    def pause_writing(self) -> None:  # pragma: no cover
        assert self._endpoint._write_ready_future is None
        assert self._endpoint._transport._loop
        loop = self._endpoint._transport._loop
        self._endpoint._write_ready_future = loop.create_future()

    def resume_writing(self) -> None:  # pragma: no cover
        assert self._endpoint._write_ready_future is not None
        self._endpoint._write_ready_future.set_result(None)
        self._endpoint._write_ready_future = None


# Enpoint classes


class Endpoint:
    """High-level interface for UDP enpoints.

    Can either be local or remote.
    It is initialized with an optional queue size for the incoming datagrams.
    """

    def __init__(self, queue_size: int | None = None) -> None:
        if queue_size is None:
            queue_size = 0
        self._queue: asyncio.Queue = asyncio.Queue(queue_size)
        self._closed = False
        self._transport: Any = None
        self._write_ready_future: asyncio.Future | None = None

    # Protocol callbacks

    def feed_datagram(self, data: bytes | None, addr: tuple[str, int] | None) -> None:
        """Feed a datagram."""
        try:
            self._queue.put_nowait((data, addr))
        except asyncio.QueueFull:
            warnings.warn("Endpoint queue is full")

    def close(self) -> None:
        """Close the endpoint."""
        # Manage flag
        if self._closed:
            return
        self._closed = True
        # Wake up
        if self._queue.empty():
            self.feed_datagram(None, None)
        # Close transport
        if self._transport:
            self._transport.close()

    # User methods

    def send(self, data: bytes, addr: tuple[str, int] | None) -> None:
        """Send a datagram to the given address."""
        if self._closed:
            raise IOError("Enpoint is closed")
        self._transport.sendto(data, addr)

    async def receive(self) -> tuple[bytes, tuple[str, int]]:
        """Wait for an incoming datagram and return it with the corresponding address.

        This method is a coroutine.
        """
        if self._queue.empty() and self._closed:
            raise IOError("Enpoint is closed")
        data, addr = await self._queue.get()
        if data is None:
            raise IOError("Enpoint is closed")
        return data, addr

    def abort(self) -> None:
        """Close the transport immediately."""
        if self._closed:
            raise IOError("Enpoint is closed")
        self._transport.abort()
        self.close()

    async def drain(self) -> None:
        """Drain the transport buffer below the low-water mark."""
        if self._write_ready_future is not None:  # pragma: no cover
            await self._write_ready_future

    # Properties

    @property
    def address(self) -> Any:
        """Endpoint address as a (host, port) tuple."""
        return self._transport.get_extra_info("socket").getsockname()

    @property
    def closed(self) -> bool:
        """Indicate whether the endpoint is closed or not."""
        return self._closed


class LocalEndpoint(Endpoint):
    """High-level interface for UDP local enpoints.

    It is initialized with an optional queue size for the incoming datagrams.
    """


class RemoteEndpoint(Endpoint):
    """High-level interface for UDP remote enpoints.

    It is initialized with an optional queue size for the incoming datagrams.
    """

    def send(self, data: bytes, addr: tuple[str, int] | None = None) -> None:
        """Send a datagram to the remote host."""
        super().send(data, None)


# High-level coroutines


async def open_datagram_endpoint(
    host: str,
    port: int,
    *,
    endpoint_factory: Any = Endpoint,
    remote: bool = False,
    **kwargs: Any
) -> Any:
    """Open and return a datagram endpoint.

    The default endpoint factory is the Endpoint class.
    The endpoint can be made local or remote using the remote argument.
    Extra keyword arguments are forwarded to `loop.create_datagram_endpoint`.
    """
    loop = asyncio.get_event_loop()
    endpoint = endpoint_factory()
    kwargs["remote_addr" if remote else "local_addr"] = host, port
    kwargs["protocol_factory"] = lambda: DatagramEndpointProtocol(endpoint)
    await loop.create_datagram_endpoint(**kwargs)
    return endpoint


async def open_local_endpoint(
    host: str = "0.0.0.0",
    port: int = 0,
    *,
    queue_size: int | None = None,
    **kwargs: dict[str, Any]
) -> Any:
    """Open and return a local datagram endpoint.

    An optional queue size arguement can be provided.
    Extra keyword arguments are forwarded to `loop.create_datagram_endpoint`.
    """
    return await open_datagram_endpoint(
        host,
        port,
        remote=False,
        endpoint_factory=lambda: LocalEndpoint(queue_size),
        **kwargs
    )


async def open_remote_endpoint(
    host: str, port: int, *, queue_size: int | None = None, **kwargs: dict[str, Any]
) -> Any:
    """Open and return a remote datagram endpoint.

    An optional queue size arguement can be provided.
    Extra keyword arguments are forwarded to `loop.create_datagram_endpoint`.
    """
    return await open_datagram_endpoint(
        host,
        port,
        remote=True,
        endpoint_factory=lambda: RemoteEndpoint(queue_size),
        **kwargs
    )



"""WeatherFlow client."""
from __future__ import annotations

import asyncio
import json
import logging
from typing import Any

from .aioudp import LocalEndpoint, open_local_endpoint
from .const import DEFAULT_HOST, DEFAULT_PORT
from .device import (
    HubDevice,
    WeatherFlowDevice,
    WeatherFlowSensorDevice,
    determine_device,
)
from .errors import AddressInUseError, EndpointError
from .mixins import EventMixin

DATA_HUB_SN = "hub_sn"
DATA_SERIAL_NUMBER = "serial_number"

EVENT_DEVICE_DISCOVERED = "device_discovered"

_LOGGER = logging.getLogger(__name__)


class WeatherFlowListener(EventMixin):
    """WeatherFlow listener."""

    def __init__(
        self,
        host: str = DEFAULT_HOST,
        port: int = DEFAULT_PORT,
    ) -> None:
        """Initialize a WeatherFlow listener."""
        self._host = host
        self._port = port
        self._devices: dict[str, WeatherFlowDevice] = {}
        self._udp_task: asyncio.Task | None = None
        self._udp_connection: LocalEndpoint | None = None

    @property
    def devices(self) -> list[WeatherFlowDevice]:
        """Return the known devices."""
        return list(self._devices.values())

    @property
    def hubs(self) -> list[HubDevice]:
        """Return the known hubs."""
        return [
            device for device in self._devices.values() if isinstance(device, HubDevice)
        ]

    @property
    def is_listening(self) -> bool:
        """Return `True` if the client is listening for messages."""
        return self._udp_task is not None

    @property
    def sensors(self) -> list[WeatherFlowSensorDevice]:
        """Return the known sensors."""
        return [
            device
            for device in self._devices.values()
            if isinstance(device, WeatherFlowSensorDevice)
        ]

    async def start_listening(self) -> None:
        """Connect the UDP socket and start listening for messages."""
        if self._udp_task is not None:
            return

        try:
            self._udp_connection = await open_local_endpoint(
                host=self._host, port=self._port
            )
        except OSError as ex:
            if "Address already in use" in ex.args:
                raise AddressInUseError("Address already in use") from ex
            raise EndpointError("Could not open a local UDP endpoint") from ex
        _LOGGER.debug("Started listening")
        self._udp_task = asyncio.ensure_future(self._start_socketreader())

    async def stop_listening(self) -> None:
        """Disconnect the socket."""
        if self._udp_task is None:
            return

        if self._udp_connection is not None and not self._udp_connection.closed:
            self._udp_connection.close()

        self._udp_task.cancel()
        try:
            await self._udp_task
        except asyncio.CancelledError:
            pass
        finally:
            self._udp_task = None
            _LOGGER.debug("Stopped listening")

    async def _start_socketreader(self) -> None:
        """Start the UDP socket listener."""
        assert self._udp_connection
        while not self._udp_connection.closed:
            data, (host, port) = await self._udp_connection.receive()
            _LOGGER.debug("Received message from %s:%s - %s", host, port, data)
            self._process_message(data)

    def _process_message(self, data: bytes) -> None:
        """Process a UDP message."""
        try:
            json_data: dict[str, Any] = json.loads(data)
            serial_number = json_data[DATA_SERIAL_NUMBER]
        except (json.JSONDecodeError, KeyError):
            _LOGGER.warning("Received unknown message: %s", data)
            return

        if serial_number not in self._devices:
            self._devices[serial_number] = determine_device(serial_number)(
                serial_number=serial_number, data=json_data
            )
            self.emit(EVENT_DEVICE_DISCOVERED, self._devices[serial_number])
        self._devices[serial_number].parse_message(json_data)

    async def __aenter__(self) -> WeatherFlowListener:
        """Connect the UDP socket and start listening for messages."""
        await self.start_listening()
        return self

    async def __aexit__(self, *exctype: Any) -> None:
        """Disconnect the socket."""
        await self.stop_listening()



"""WeatherFlow constants."""
import functools
import re

import pint

units = pint.UnitRegistry(
    autoconvert_offset_to_baseunit=True,
    preprocessors=[
        functools.partial(
            re.sub,
            r"(?<=[A-Za-z])(?![A-Za-z])(?<![0-9\-][eE])(?<![0-9\-])(?=[0-9\-])",
            "**",
        ),
        lambda string: string.replace("%", "percent"),
    ],
)
units.define(
    pint.unit.UnitDefinition("percent", "%", (), pint.converters.ScaleConverter(0.01))
)
units.default_format = "P~"

DEFAULT_HOST = "0.0.0.0"
DEFAULT_PORT = 50222

EVENT_OBSERVATION_AIR = "obs_air"
EVENT_OBSERVATION_SKY = "obs_sky"
EVENT_OBSERVATION_TEMPEST = "obs_st"
EVENT_RAIN_START = "evt_precip"
EVENT_RAPID_WIND = "rapid_wind"
EVENT_STATUS_DEVICE = "device_status"
EVENT_STATUS_HUB = "hub_status"
EVENT_STRIKE = "evt_strike"

UNIT_DECIBELS = units.dB
UNIT_DEGREES = units.deg
UNIT_DEGREES_CELSIUS = units.degC
UNIT_IRRADIATION = units.W / units.m ** 2
UNIT_KILOGRAMS_PER_CUBIC_METER = units.kg / units.m ** 3
UNIT_KILOMETERS = units.km
UNIT_LUX = units.lx
UNIT_METERS_PER_SECOND = units.m / units.sec
UNIT_MILLIBARS = units.mbar
UNIT_MILLIMETERS = units.mm
UNIT_MILLIMETERS_PER_MINUTE = units.mm / units.min
UNIT_MILLIMETERS_PER_HOUR = units.mm / units.hr
UNIT_MINUTES = units.min
UNIT_PERCENT = units.percent
UNIT_SECONDS = units.sec
UNIT_VOLTS = units.V



"""Errors for WeatherFlow UDP connections."""


class WeatherflowUdpError(Exception):
    """Base WeatherFlow UDP error."""


class ListenerError(WeatherflowUdpError):
    """Error indicating a proplem with the listener."""


class AddressInUseError(ListenerError):
    """Error indicating the listener address is already in use."""


class EndpointError(ListenerError):
    """Error indicating an issue with the UDP endpoint."""



"""Helper functions."""
from __future__ import annotations

from datetime import datetime, timezone
from typing import Any

from pint import Quantity

DIRECTIONS = [
    "N",
    "NNE",
    "NE",
    "ENE",
    "E",
    "ESE",
    "SE",
    "SSE",
    "S",
    "SSW",
    "SW",
    "WSW",
    "W",
    "WNW",
    "NW",
    "NNW",
]
UTC = timezone.utc


def degrees_to_cardinal(degree: float | Quantity[float]) -> str:
    """Convert degrees to a cardinal direction."""
    return DIRECTIONS[
        round(
            (degree.m if isinstance(degree, Quantity) else degree)
            / (360.0 / (directions_count := len(DIRECTIONS)))
        )
        % directions_count
    ]


def truebool(val: Any | None) -> bool:
    """Return `True` if the value passed in matches a "True" value, otherwise `False`.

    "True" values are: 'true', 't', 'yes', 'y', 'on' or '1'.
    """
    return val is not None and str(val).lower() in ("true", "t", "yes", "y", "on", "1")


def utc_timestamp_from_epoch(epoch: int | None) -> datetime | None:
    """Return the UTC timestamp from an epoch value."""
    return None if epoch is None else datetime.fromtimestamp(epoch, UTC)



"""Calc functions."""
from __future__ import annotations

import psychrolib
from pint import Quantity

from .const import (
    UNIT_DEGREES_CELSIUS,
    UNIT_KILOGRAMS_PER_CUBIC_METER,
    UNIT_MILLIBARS,
    UNIT_PERCENT,
    units,
)

psychrolib.SetUnitSystem(psychrolib.SI)


def air_density(
    air_temperature: Quantity[float], station_pressure: Quantity[float]
) -> Quantity[float]:
    """Calculate the air density in kilograms per cubic meter (kg/m³)."""
    return (
        psychrolib.GetDryAirDensity(
            air_temperature.to("degC").m, station_pressure.to("Pa").m
        )
        * UNIT_KILOGRAMS_PER_CUBIC_METER
    )


def dew_point_temperature(
    air_temperature: Quantity[float], relative_humidity: Quantity[float]
) -> Quantity[float]:
    """Calculate the dew point temperature."""
    return (
        psychrolib.GetTDewPointFromRelHum(
            air_temperature.to("degC").m,
            relative_humidity.m
            / (
                100
                if relative_humidity.u == UNIT_PERCENT or relative_humidity.m > 1
                else 1
            ),
        )
        * UNIT_DEGREES_CELSIUS
    ).to(air_temperature.u)


def heat_index(
    air_temperature: Quantity[float], relative_humidity: Quantity[float]
) -> Quantity[float] | None:
    """Calculate the heat index.

    Only temperatures >= 80°F (26.66°C) have a heat index.
    """
    temp_fahrenheit = air_temperature.to("degF").m

    if temp_fahrenheit < 80:
        return None

    rh_percent = relative_humidity.m * (
        1 if relative_humidity.u == UNIT_PERCENT or relative_humidity.m > 1 else 100
    )

    heat_idx = 0.5 * (
        temp_fahrenheit + 61.0 + ((temp_fahrenheit - 68.0) * 1.2) + (rh_percent * 0.094)
    )
    if (heat_idx + temp_fahrenheit) / 2 >= 80:
        heat_idx = (
            -42.379
            + 2.04901523 * temp_fahrenheit
            + 10.14333127 * rh_percent
            - 0.22475541 * temp_fahrenheit * rh_percent
            - 0.00683783 * temp_fahrenheit * temp_fahrenheit
            - 0.05481717 * rh_percent * rh_percent
            + 0.00122874 * temp_fahrenheit * temp_fahrenheit * rh_percent
            + 0.00085282 * temp_fahrenheit * rh_percent * rh_percent
            - 0.00000199 * temp_fahrenheit * temp_fahrenheit * rh_percent * rh_percent
        )
        if rh_percent < 13 and (80 <= temp_fahrenheit <= 112):
            heat_idx -= ((13 - rh_percent) / 4) * (
                (17 - abs(temp_fahrenheit - 95.0)) / 17
            ) ** 0.5
        elif rh_percent > 85 and (80 <= temp_fahrenheit <= 87):
            heat_idx += ((rh_percent - 85) / 10) * ((87 - temp_fahrenheit) / 5)
    return (heat_idx * units.degF).to(air_temperature.u)


def sea_level_pressure(
    station_pressure: Quantity[float],
    altitude: Quantity[float],
    air_temperature: Quantity[float],
) -> Quantity[float]:
    """Calculate the sea level pressure."""
    return (
        psychrolib.GetSeaLevelPressure(
            station_pressure.to("Pa").m,
            altitude.to("m").m,
            air_temperature.to("degC").m,
        )
        * units.Pa
    ).to(station_pressure.u)


def vapor_pressure(
    air_temperature: Quantity[float], relative_humidity: Quantity[float]
) -> Quantity[float]:
    """Calculate the vapor pressure in millibars (mbar)."""
    return (
        psychrolib.GetVapPresFromRelHum(
            air_temperature.to("degC").m,
            relative_humidity.m
            / (
                100
                if relative_humidity.u == UNIT_PERCENT or relative_humidity.m > 1
                else 1
            ),
        )
        * UNIT_MILLIBARS
    )


def wet_bulb_temperature(
    air_temperature: Quantity[float],
    relative_humidity: Quantity[float],
    station_pressure: Quantity[float],
) -> Quantity[float]:
    """Calculate the wet bulb temperature."""
    return (
        psychrolib.GetTWetBulbFromRelHum(
            air_temperature.to("degC").m,
            relative_humidity.m / (100 if relative_humidity.m > 1 else 1),
            station_pressure.to("Pa").m,
        )
        * UNIT_DEGREES_CELSIUS
    ).to(air_temperature.u)


def wind_chill(
    air_temperature: Quantity[float], wind_speed: Quantity[float]
) -> Quantity[float] | None:
    """Calculate the wind chill temperature.

    Only temperatures <= 50°F (10°C) and winds >= 3mph (1.34112 m/s) have a wind chill.
    """
    temp_fahrenheit = air_temperature.to("degF").m
    wind_mph = wind_speed.to("mph").m

    if temp_fahrenheit > 50 or wind_mph < 3:
        return None

    return (
        (
            35.74
            + (0.6215 * temp_fahrenheit)
            - (35.75 * pow(wind_mph, 0.16))
            + (0.4275 * temp_fahrenheit * pow(wind_mph, 0.16))
        )
        * units.degF
    ).to(air_temperature.u)



"""WeatherFlow devices."""
from __future__ import annotations

import logging
from datetime import datetime
from typing import Any, Callable, final

from pint import Quantity

from .calc import wind_chill
from .const import (
    EVENT_OBSERVATION_AIR,
    EVENT_OBSERVATION_SKY,
    EVENT_OBSERVATION_TEMPEST,
    EVENT_RAIN_START,
    EVENT_RAPID_WIND,
    EVENT_STATUS_DEVICE,
    EVENT_STATUS_HUB,
    EVENT_STRIKE,
    UNIT_DECIBELS,
    UNIT_SECONDS,
)
from .event import CustomEvent, LightningStrikeEvent, RainStartEvent, WindEvent
from .helpers import truebool, utc_timestamp_from_epoch
from .mixins import AirSensorMixin, BaseSensorMixin, EventMixin, SkySensorMixin

DATA_DEBUG = "debug"
DATA_EVENT = "evt"
DATA_FIRMWARE_REVISION = "firmware_revision"
DATA_HUB_RSSI = "hub_rssi"
DATA_HUB_SN = "hub_sn"
DATA_OBSERVATION = "ob"
DATA_OBSERVATIONS = "obs"
DATA_RADIO_STATS = "radio_stats"
DATA_RESET_FLAGS = "reset_flags"
DATA_RSSI = "rssi"
DATA_SENSOR_STATUS = "sensor_status"
DATA_SEQ = "seq"
DATA_TIMESTAMP = "timestamp"
DATA_TYPE = "type"
DATA_UPTIME = "uptime"
DATA_VOLTAGE = "voltage"

EVENT_LOAD_COMPLETE = "load_complete"
EVENT_OBSERVATION = "observation"
EVENT_STATUS_UPDATE = "status_update"

RESET_FLAGS = {
    "BOR": "Brownout reset",
    "PIN": "PIN reset",
    "POR": "Power reset",
    "SFT": "Software reset",
    "WDG": "Watchdog reset",
    "WWD": "Window watchdog reset",
    "LPW": "Low-power reset",
    "HRDFLT": "Hard fault detected",
}

RAIN_START_EVENT_TIMESTAMP = 0

STRIKE_EVENT_TIMESTAMP = 0
STRIKE_EVENT_DISTANCE = 1
STRIKE_EVENT_ENERGY = 2

WIND_OBSERVATION_TIMESTAMP = 0
WIND_OBSERVATION_SPEED = 1
WIND_OBSERVATION_DIRECTION = 2

_LOGGER = logging.getLogger(__name__)


class WeatherFlowDevice(EventMixin):
    """Generic WeatherFlow device."""

    _attr_model = "Unknown"

    def __init__(self, serial_number: str, data: dict[str, Any] | None = None) -> None:
        """Initialize a WeatherFlow device."""
        # pylint: disable=unused-argument
        self._serial_number = serial_number
        self._firmware_revision: str | None = None
        self._rssi: int = 0
        self._timestamp: int | None = None
        self._uptime: int = 0
        self._initial_status: bool = False

        self._listeners: dict[str, list[Callable]] = {}
        self._parse_message_map: dict[str, Callable | tuple[Callable, str]] = {}

    @property
    def firmware_revision(self) -> str:
        """Return firmware revision."""
        return str(self._firmware_revision)

    @property
    def load_complete(self) -> bool:
        """Return `True` if the device has been completely loaded."""
        return self._initial_status

    @property
    def model(self) -> str:
        """Return the model."""
        return self._attr_model

    @property
    def rssi(self) -> Quantity[int]:
        """Return the rssi."""
        return self._rssi * UNIT_DECIBELS

    @property
    def serial_number(self) -> str:
        """Return the serial number."""
        return self._serial_number

    @property
    def timestamp(self) -> datetime | None:
        """Return the timestamp in UTC."""
        return utc_timestamp_from_epoch(self._timestamp)

    @property
    def up_since(self) -> datetime | None:
        """Return the device's up since timestamp in UTC."""
        return (
            None
            if self._timestamp is None
            else utc_timestamp_from_epoch(self._timestamp - self._uptime)
        )

    @property
    def uptime(self) -> Quantity[int]:
        """Return the uptime in seconds."""
        return self._uptime * UNIT_SECONDS

    def register_parse_handlers(
        self, message_handlers: dict[str, Callable | tuple[Callable, str]]
    ) -> None:
        """Register or replace a parse handler."""
        for message_type, handler in message_handlers.items():
            self._parse_message_map[message_type] = handler

    @final
    def parse_message(self, data: dict[str, Any]) -> None:
        """Parse a device message."""
        if handler := self._parse_message_map.get(
            message_type := data.get(DATA_TYPE, "")
        ):
            if isinstance(handler, tuple):
                handler, field = handler
                data = data.get(field, {})
            handler(data)
        else:
            _LOGGER.warning("Unhandled %s message: %s", message_type, data)


class HubDevice(WeatherFlowDevice):
    """Represents a WeatherFlow Hub."""

    _attr_model = "Hub"

    def __init__(self, serial_number: str, data: dict[str, Any] | None = None) -> None:
        """Initialize a WeatherFlow hub device."""
        super().__init__(serial_number=serial_number, data=data)

        self._reset_flags: str = ""
        self._seq: int = 0
        self._radio_stats: list[int] = []

        self.register_parse_handlers({EVENT_STATUS_HUB: self.parse_hub_status})

    def __repr__(self) -> str:  # pragma: no cover
        """Return repr(self)."""
        return f"{self.model}<serial_number={self.serial_number}>"

    @property
    def reset_flags(self) -> list[str] | None:
        """Return the list of reset flags."""
        return [
            RESET_FLAGS[flag]
            for flag in self._reset_flags.split(",")
            if flag in RESET_FLAGS
        ]

    def parse_hub_status(self, data: dict[str, Any]) -> None:
        """Parse hub status."""
        self._firmware_revision = data.get(DATA_FIRMWARE_REVISION)
        self._uptime = data.get(DATA_UPTIME, 0)
        self._rssi = data.get(DATA_RSSI, 0)
        self._timestamp = data.get(DATA_TIMESTAMP)
        self._reset_flags = data.get(DATA_RESET_FLAGS, "")
        self._seq = data.get(DATA_SEQ, 0)
        self._radio_stats = data.get(DATA_RADIO_STATS, [])

        assert self._timestamp

        if not self._initial_status:
            self._initial_status = True
            self.emit(
                EVENT_LOAD_COMPLETE,
                CustomEvent(self._timestamp, EVENT_LOAD_COMPLETE),
            )

        self.emit(
            EVENT_STATUS_UPDATE, CustomEvent(self._timestamp, EVENT_STATUS_UPDATE)
        )


class WeatherFlowSensorDevice(BaseSensorMixin, WeatherFlowDevice):
    """Represents a WeatherFlow sensor device."""

    _evt_observation: str = ""

    OBSERVATION_VALUES_MAP: dict[int, str] = {}

    SENSOR_STATUS_MASK = {
        0b000000001: "Lightning Failed",
        0b000000010: "Lightning Noise",
        0b000000100: "Lightning Disturber",
        0b000001000: "Pressure Failed",
        0b000010000: "Temperature Failed",
        0b000100000: "Relative Humidity Failed",
        0b001000000: "Wind Failed",
        0b010000000: "Precipitation Failed",
        0b100000000: "Light/UV Failed",
        0x00008000: "Power Booster Depleted",
        0x00010000: "Power Booster Shore Power",
    }

    def __init__(self, serial_number: str, data: dict[str, Any] | None = None) -> None:
        """Initialize a WeatherFlow device."""
        super().__init__(serial_number=serial_number, data=data)

        self._hub_sn: str | None = data.get(DATA_HUB_SN) if data is not None else None
        self._voltage: float = 0
        self._hub_rssi: int = 0
        self._sensor_status: int = 0
        self._debug: bool = False
        self._initial_observation: bool = False

        self.register_parse_handlers(
            {
                self._evt_observation: (self.parse_observation, DATA_OBSERVATIONS),
                EVENT_STATUS_DEVICE: self.parse_device_status,
            }
        )

    def __repr__(self) -> str:  # pragma: no cover
        """Return repr(self)."""
        return f"{self.model}<serial_number={self.serial_number}, hub={self.hub_sn}>"

    @property
    def hub_rssi(self) -> Quantity[int]:
        """Return the hub rssi."""
        return self._hub_rssi * UNIT_DECIBELS

    @property
    def hub_sn(self) -> str | None:
        """Return the hub serial number."""
        return self._hub_sn

    @property
    def load_complete(self) -> bool:
        """Return `True` if the device has been completely loaded."""
        return self._initial_status and self._initial_observation

    @property
    def sensor_status(self) -> list[str]:
        """Return the sensor status."""
        return (
            []
            if self._sensor_status is None
            else [
                status
                for mask in self.SENSOR_STATUS_MASK
                if (status := self.SENSOR_STATUS_MASK.get(mask & self._sensor_status))
                is not None
            ]
        )

    def parse_device_status(self, data: dict[str, Any]) -> None:
        """Parse the device status."""
        self._timestamp = data.get(DATA_TIMESTAMP)
        self._uptime = data.get(DATA_UPTIME, 0)
        self._voltage = data.get(DATA_VOLTAGE, 0)
        self._firmware_revision = data.get(DATA_FIRMWARE_REVISION)
        self._rssi = data.get(DATA_RSSI, 0)
        self._hub_rssi = data.get(DATA_HUB_RSSI, 0)
        self._sensor_status = data.get(DATA_SENSOR_STATUS, 0)
        self._debug = truebool(data.get(DATA_DEBUG))

        if not self._initial_status:
            self._initial_status = True
            self._send_load_complete_event()

        assert self._timestamp
        self.emit(
            EVENT_STATUS_UPDATE, CustomEvent(self._timestamp, EVENT_STATUS_UPDATE)
        )

    def parse_observation(self, data: list[list]) -> None:
        """Parse observation data."""
        for observation in data:
            for idx, field in self.OBSERVATION_VALUES_MAP.items():
                setattr(self, field, observation[idx])

        if not self._initial_observation:
            self._initial_observation = True
            self._send_load_complete_event()

        assert self._last_report
        self.emit(EVENT_OBSERVATION, CustomEvent(self._last_report, EVENT_OBSERVATION))

    def _send_load_complete_event(self) -> None:
        if self._initial_status and self._initial_observation:
            assert self._last_report
            assert self._timestamp
            self.emit(
                EVENT_LOAD_COMPLETE,
                CustomEvent(
                    max(self._timestamp, self._last_report), EVENT_LOAD_COMPLETE
                ),
            )


class AirSensorType(AirSensorMixin, WeatherFlowSensorDevice):
    """Air sensor type class."""

    def __init__(self, serial_number: str, data: dict[str, Any] | None = None) -> None:
        """Initialize a WeatherFlow air sensor device."""
        super().__init__(serial_number=serial_number, data=data)

        self.register_parse_handlers(
            {EVENT_STRIKE: (self.parse_strike_event, DATA_EVENT)}
        )

    def parse_strike_event(self, data: list) -> None:
        """Parse strike event data."""
        if (report_timestamp := data[STRIKE_EVENT_TIMESTAMP]) < (
            0
            if self._last_lightning_strike_event is None
            else self._last_lightning_strike_event.epoch
        ):  # pragma: no cover
            _LOGGER.warning("Received an old strike event: %s", data)
            return

        self._last_lightning_strike_event = LightningStrikeEvent(
            report_timestamp, data[STRIKE_EVENT_DISTANCE], data[STRIKE_EVENT_ENERGY]
        )

        self.emit(EVENT_STRIKE, self._last_lightning_strike_event)


class SkySensorType(SkySensorMixin, WeatherFlowSensorDevice):
    """Sky sensor type class."""

    def __init__(self, serial_number: str, data: dict[str, Any] | None = None) -> None:
        """Initialize a WeatherFlow device."""
        super().__init__(serial_number=serial_number, data=data)

        self.register_parse_handlers(
            {
                EVENT_RAIN_START: (self.parse_rain_start_event, DATA_EVENT),
                EVENT_RAPID_WIND: (self.parse_wind_event, DATA_OBSERVATION),
            }
        )

    def parse_rain_start_event(self, data: list) -> None:
        """Parse rain start event data."""
        if (report_timestamp := data[RAIN_START_EVENT_TIMESTAMP]) < (
            0
            if self._last_rain_start_event is None
            else self._last_rain_start_event.epoch
        ):  # pragma: no cover
            _LOGGER.warning("Received an old rain start event: %s", data)
            return

        self._last_rain_start_event = RainStartEvent(report_timestamp)

        self.emit(EVENT_RAIN_START, self._last_rain_start_event)

    def parse_wind_event(self, data: list) -> None:
        """Parse wind event data."""
        if (report_timestamp := data[WIND_OBSERVATION_TIMESTAMP]) < (
            0 if self._last_wind_event is None else self._last_wind_event.epoch
        ):  # pragma: no cover
            _LOGGER.warning("Received an old wind report: %s", data)
            return

        self._last_wind_event = WindEvent(
            report_timestamp,
            data[WIND_OBSERVATION_SPEED],
            data[WIND_OBSERVATION_DIRECTION],
        )

        self.emit(EVENT_RAPID_WIND, self._last_wind_event)


class AirDevice(AirSensorType):
    """Represents a WeatherFlow Air device."""

    _attr_model = "Air"
    _evt_observation = EVENT_OBSERVATION_AIR

    OBSERVATION_VALUES_MAP = {
        0: "_last_report",
        1: "_station_pressure",
        2: "_air_temperature",
        3: "_relative_humidity",
        4: "_lightning_strike_count",
        5: "_lightning_strike_average_distance",
        6: "_battery",
        7: "_report_interval",
    }


class SkyDevice(SkySensorType):
    """Represents a WeatherFlow Sky device."""

    _attr_model = "Sky"
    _evt_observation = EVENT_OBSERVATION_SKY

    OBSERVATION_VALUES_MAP = {
        0: "_last_report",
        1: "_illuminance",
        2: "_uv",
        3: "_rain_accumulation_previous_minute",
        4: "_wind_lull",
        5: "_wind_average",
        6: "_wind_gust",
        7: "_wind_direction",
        8: "_battery",
        9: "_report_interval",
        10: "_solar_radiation",
        12: "_precipitation_type",
        13: "_wind_sample_interval",
    }


class TempestDevice(AirSensorType, SkySensorType):
    """Represents a WeatherFlow Tempest device."""

    _attr_model = "Tempest"
    _evt_observation = EVENT_OBSERVATION_TEMPEST

    OBSERVATION_VALUES_MAP = {
        0: "_last_report",
        1: "_wind_lull",
        2: "_wind_average",
        3: "_wind_gust",
        4: "_wind_direction",
        5: "_wind_sample_interval",
        6: "_station_pressure",
        7: "_air_temperature",
        8: "_relative_humidity",
        9: "_illuminance",
        10: "_uv",
        11: "_solar_radiation",
        12: "_rain_accumulation_previous_minute",
        13: "_precipitation_type",
        14: "_lightning_strike_average_distance",
        15: "_lightning_strike_count",
        16: "_battery",
        17: "_report_interval",
    }

    # Derived metrics

    @property
    def feels_like_temperature(self) -> Quantity[float]:
        """Return the feels like temperature in degrees Celsius (°C)."""
        if self.heat_index is not None:
            return self.heat_index
        if self.wind_chill_temperature is not None:
            return self.wind_chill_temperature
        return self.air_temperature

    @property
    def wind_chill_temperature(self) -> Quantity[float]:
        """Return the calculated wind chill temperature in degrees Celsius (°C)."""
        return wind_chill(self.air_temperature, self.wind_speed)


SERIAL_MAP = {"HB": HubDevice, "AR": AirDevice, "SK": SkyDevice, "ST": TempestDevice}


def determine_device(serial_number: str) -> type[WeatherFlowDevice]:
    """Return the type of WeatherFlow device from a serial number."""
    return SERIAL_MAP.get(serial_number[:2], WeatherFlowDevice)



"""Events for WeatherFlow devices."""
from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime

from pint import Quantity

from .const import UNIT_DEGREES, UNIT_KILOMETERS, UNIT_METERS_PER_SECOND
from .helpers import utc_timestamp_from_epoch

# pylint: disable=line-too-long


@dataclass
class Event:
    """Event base class."""

    _epoch: int

    @property
    def epoch(self) -> int:
        """Return epoch in seconds."""
        return self._epoch

    @property
    def timestamp(self) -> datetime | None:
        """Return the timestamp in UTC."""
        return utc_timestamp_from_epoch(self._epoch)

    def __repr__(self) -> str:  # pragma: no cover
        """Return repr(self)."""
        return f"Event<timestamp={self.timestamp}>"


@dataclass
class LightningStrikeEvent(Event):
    """Lightning strike event class."""

    _distance: float
    _energy: int

    @property
    def distance(self) -> Quantity[float]:
        """Return the distance in kilometers."""
        return self._distance * UNIT_KILOMETERS

    @property
    def energy(self) -> int:
        """Return the energy.

        Energy is just a pure number and has no physical meaning.
        """
        return self._energy

    def __repr__(self) -> str:  # pragma: no cover
        """Return repr(self)."""
        return (
            f"Lightning Strike Event<timestamp={self.timestamp}, speed={self.distance}>"
        )


@dataclass
class RainStartEvent(Event):
    """Rain start event class."""

    def __repr__(self) -> str:  # pragma: no cover
        """Return repr(self)."""
        return f"Rain Start Event<timestamp={self.timestamp}>"


@dataclass
class WindEvent(Event):
    """Wind event class."""

    _speed: float
    _direction: int

    @property
    def direction(self) -> Quantity[int]:
        """Return the direction in degrees."""
        return self._direction * UNIT_DEGREES

    @property
    def speed(self) -> Quantity[float]:
        """Return the speed in meters per second."""
        return self._speed * UNIT_METERS_PER_SECOND

    def __repr__(self) -> str:  # pragma: no cover
        """Return repr(self)."""
        return f"Wind Event<timestamp={self.timestamp}, speed={self.speed}, direction={self.direction.m}°>"


@dataclass
class CustomEvent(Event):
    """Custom event."""

    name: str

    def __repr__(self) -> str:  # pragma: no cover
        """Return repr(self)."""
        return f"Event<timestamp={self.timestamp}, name={self.name}>"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyweatherflowudp']

package_data = \
{'': ['*']}

install_requires = \
['Pint>=0.18,<0.19', 'PsychroLib>=2.5.0,<3.0.0']

setup_kwargs = {
    'name': 'pyweatherflowudp',
    'version': '1.1.1',
    'description': 'An event-based asynchronous library to read UDP packets from Weatherflow weather systems on a local network without any reliance on the cloud.',
    'long_description': '# pyweatherflowudp\n\nAn event-based asynchronous library to read UDP packets from Weatherflow weather systems on a local network without any reliance on the cloud.\n\nThe atomic data provided by WeatherFlow sensors is all natively metric. To help facilitate transformations to other units or to perform calculations and comparisons, this module utilizes [Pint](https://pint.readthedocs.io/en/stable/)\'s Quantity class as the type of most properties which have a unit of measurement. See the [Quantity](#Quantity) section below for more details.\n\nThis module utilizes [PsychroLib](https://github.com/psychrometrics/psychrolib) to help with additional weather calculations that are derived from the various data points provided by the actual WeatherFlow sensors. While WeatherFlow has these additional data points available via the app and rest API, they are unavailable via the low-level UDP packet data. And while a list of [derived metrics](https://weatherflow.github.io/Tempest/api/derived-metric-formulas.html) and the associated equations has been posted, they can be quite complex to implement, such as calculating wet bulb temperature. As such, PsychroLib is an invaluable resource since the work has already been done and it eliminates the need to write all the equations in this module (and potentially get them wrong). You may notice that some of these values aren\'t an exact match with what is shown in the WeatherFlow app because there are different formulas (some simpler, some more complex) to calculate derived weather metrics. This is because WeatherFlow and PsychroLib may have chosen one of the sometimes many different formulas to get the result desired. They should still be relatively close, however.\n\n[![DOI](https://joss.theoj.org/papers/10.21105/joss.01137/status.svg)](https://doi.org/10.21105/joss.01137) [![DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.2537945.svg)](https://doi.org/10.5281/zenodo.2537945)\n\n## Supported Devices\n\n- Tempest\n- Air (untested)\n- Sky (untested)\n\n## Usage\n\nThe primary way of interacting with this module, and the devices that are found with it, is via event subscriptions.\n\nThe classes and events which they currently support are outlined below, but you can also see an example tying this altogether in [demo.py](/demo.py).\n\n### Client\n\n#### `WeatherFlowListener`\n\n- EVENT_DEVICE_DISCOVERED - emitted when a new device (Hub, Air, Sky or Tempest) is disovered for the first time\n\nThe classes and events in this section can be imported from `pyweatherflowudp.client`.\n\n### Devices\n\n#### `HubDevice`\n\n- EVENT_LOAD_COMPLETE - emitted once when the hub has finished parsing a status update for the first time\n- EVENT_STATUS_UPDATE - emitted each time a status update has been parsed by the hub, including immediately following an EVENT_LOAD_COMPLETE\n\n#### `WeatherFlowSensorDevice` (base class for `AirDevice`, `SkyDevice` and `TempestDevice`)\n\n- EVENT_LOAD_COMPLETE - emitted once when the sensor has finished parsing one status update and one observation for the first time\n- EVENT_OBSERVATION - emitted each time an observation has been parsed by the sensor\n- EVENT_STATUS_UPDATE - emitted each time a status update has been parsed by the sensor\n\n#### `AirSensorType` (`AirDevice` / `TempestDevice`)\n\n- EVENT_STRIKE - emitted when a lightning strike has been detected\n\n#### `SkySensorType` (`SkyDevice` / `TempestDevice`)\n\n- EVENT_RAIN_START - emitted when rain has been detected\n- EVENT_RAPID_WIND - emitted every 3 seconds (or less frequently depending on battery voltage) to inform about current wind condtions\n\nThe classes and events in this section can be imported from `pyweatherflowudp.device`.\n\n## Properties\n\n### WeatherFlowListener\n\n| property     | type | description                                                                              |\n| ------------ | ---- | ---------------------------------------------------------------------------------------- |\n| devices      | list | The known devices.                                                                       |\n| hubs         | list | The known hubs.                                                                          |\n| is_listening | bool | `True` if the listener is currently monitoring UDP packets on the network, else `False`. |\n| sensors      | list | The known sensors.                                                                       |\n\n### WeatherFlowDevice\n\nBase for hubs and sensors.\n\n| property          | type     | description                                                         |\n| ----------------- | -------- | ------------------------------------------------------------------- |\n| firmware_revision | str      | The current firmware revision of the device.                        |\n| load_complete     | bool     | `True` if the device has parsed all initial updates, else `False`.  |\n| model             | str      | The model of the device ("Hub", "Air", "Sky", "Tempest").           |\n| rssi              | Quantity | The signal strength of the device in decibels.                      |\n| serial_number     | str      | The serial number of the device.                                    |\n| timestamp         | datetime | The UTC timestamp from the last status update.                      |\n| up_since          | datetime | The UTC timestamp the device started up and has since been running. |\n| uptime            | int      | The number of seconds the device has been up and running.           |\n\n### HubDevice\n\n| property    | type | description                         |\n| ----------- | ---- | ----------------------------------- |\n| reset_flags | list | The current reset flags of the hub. |\n\n### WeatherFlowSensorDevice\n\nBase for sensors.\n\n| property        | type     | description                                           |\n| --------------- | -------- | ----------------------------------------------------- |\n| battery         | Quantity | The current battery voltage.                          |\n| hub_rssi        | Quantity | The signal strength of the hub in decibels.           |\n| hub_sn          | str      | The serial number of the hub the sensor belongs to.   |\n| last_report     | datetime | The UTC timestamp from the last observation.          |\n| sensor_status   | list     | The list of issues the sensor is currently reporting. |\n| report_interval | Quantity | The report interval in minutes.                       |\n| reset_flags     | list     | The current reset flags of the hub.                   |\n\n### AirSensorType\n\nBase for "air" sensor measurements (Air/Tempest).\n\n| property                                   | type                 | description                                                            |\n| ------------------------------------------ | -------------------- | ---------------------------------------------------------------------- |\n| air_temperature                            | Quantity             | The current air temperature in degrees Celsius.                        |\n| last_lightning_strike_event                | LightningStrikeEvent | The last lightning strike event.                                       |\n| lightning_strike_average_distance          | Quantity             | The average distance for lightning strikes in kilometers.              |\n| lightning_strike_count                     | int                  | The number of lightning strikes.                                       |\n| relative_humidity                          | Quantity             | The relative humidity percentage.                                      |\n| station_pressure                           | Quantity             | The observed station pressure in millibars.                            |\n| air_density\\*                              | Quantity             | The calculated air density in kilograms per cubic meter.               |\n| delta_t\\*                                  | Quantity             | The calculated Delta T in delta degrees Celsius.                       |\n| dew_point_temperature\\*                    | Quantity             | The calculated dew point temperature in degrees Celsius.               |\n| heat_index\\*                               | Quantity             | The calculated heat index in degrees Celsius.                          |\n| vapor_pressure\\*                           | Quantity             | The calculated vapor pressure in millibars.                            |\n| wet_bulb_temperature\\*                     | Quantity             | The calculated wet bulb temperature in degrees Celsius.                |\n| _calculate_sea_level_pressure(height)_\\*\\* | Quantity             | Calculate the sea level pressure in millibars from a specified height. |\n\n\\* Indicates derived properties\n\n\\*\\* Indicates this is a method, not a property.\n\n### SkySensorType\n\nBase for "sky" sensor measurements (Sky/Tempest).\n\n| property                          | type              | description                                                                        |\n| --------------------------------- | ----------------- | ---------------------------------------------------------------------------------- |\n| illuminance                       | Quantity          | The current illuminance in Lux.                                                    |\n| last_rain_start_event             | RainStartEvent    | The last rain start event.                                                         |\n| last_wind_event                   | WindEvent         | The last wind event.                                                               |\n| precipitation_type                | PrecipitationType | The current precipitation type: (NONE, RAIN, HAIL or RAIN_HAIL).                   |\n| rain_accumulation_previous_minute | Quantity          | The rain accumulation from the previous minute in millimeters.                     |\n| rain_rate\\*                       | Quantity          | The rain rate in millimeters per hour (based on the previous minute accumulation). |\n| solar_radiation                   | Quantity          | The solar radiation in Watts per cubic meter.                                      |\n| uv                                | int               | The current UV index.                                                              |\n| wind_average                      | Quantity          | The wind speed average over the report interval in meters per second.              |\n| wind_direction                    | Quantity          | The wind direction over the report interval in degrees.                            |\n| wind_direction_cardinal           | string            | The wind direction cardinal (16-wind compass rose).                                |\n| wind_gust                         | Quantity          | The wind gust (maximum 3 second sample) in meters per second.                      |\n| wind_lull                         | Quantity          | The wind lull (minimum 3 second sample) in meters per second.                      |\n| wind_sample_interval              | Quantity          | The wind sample interval in seconds.                                               |\n| wind_speed                        | Quantity          | The wind speed in meters per second.                                               |\n\n\\* Indicates derived properties\n\n### TempestDevice\n\n| property                 | type     | description                                                 |\n| ------------------------ | -------- | ----------------------------------------------------------- |\n| feels_like_temperature\\* | Quantity | The calculated "feels like" temperature in degrees Celsius. |\n| wind_chill_temperature\\* | Quantity | The calculated wind chill temperature in degrees Celsius.   |\n\n\\* Indicates derived properties\n\n### Quantity\n\nThe `pint.Quantity` class has been utilized for device properties which are associated with a unit of measurement. This allows a conversion from the native metric unit to another of the user\'s choice such as degrees Celsius to degrees Fahrenheit, which produces another `pint.Quantity`:\n\n```python\ndevice.air_temperature.to("degF")\n```\n\nTo retrieve only the numeric value of a property, you can just append a `.magnitude` (or `.m` short form) like:\n\n```python\ndevice.air_temperature.m\n```\n\nYou can also retrieve only the units portion of a property with `.units` (or `.u` short form) like:\n\n```python\ndevice.air_temperature.u\n```\n\nCheck out the [Pint docs](https://pint.readthedocs.io/en/stable/#user-guide) for more tips.\n',
    'author': 'Nathan Spencer',
    'author_email': 'natekspencer@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/briis/pyweatherflowudp',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Mixins for WeatherFlow client and sensors."""
from __future__ import annotations

import logging
from datetime import datetime
from typing import Any, Callable

from pint import Quantity

from .calc import (
    air_density,
    dew_point_temperature,
    heat_index,
    sea_level_pressure,
    vapor_pressure,
    wet_bulb_temperature,
)
from .const import (
    UNIT_DEGREES,
    UNIT_DEGREES_CELSIUS,
    UNIT_IRRADIATION,
    UNIT_KILOMETERS,
    UNIT_LUX,
    UNIT_METERS_PER_SECOND,
    UNIT_MILLIBARS,
    UNIT_MILLIMETERS,
    UNIT_MILLIMETERS_PER_HOUR,
    UNIT_MILLIMETERS_PER_MINUTE,
    UNIT_MINUTES,
    UNIT_PERCENT,
    UNIT_VOLTS,
)
from .enums import PrecipitationType
from .event import LightningStrikeEvent, RainStartEvent, WindEvent
from .helpers import degrees_to_cardinal, utc_timestamp_from_epoch

_LOGGER = logging.getLogger(__name__)


class EventMixin:
    """Event mixin."""

    _listeners: dict[str, list[Callable]] = {}

    def on(  # pylint: disable=invalid-name
        self, event_name: str, callback: Callable
    ) -> Callable:
        """Register an event callback."""
        listeners: list = self._listeners.setdefault(event_name, [])
        listeners.append(callback)

        def unsubscribe() -> None:
            """Unsubscribe listeners."""
            if callback in listeners:
                listeners.remove(callback)

        return unsubscribe

    def emit(self, event_name: str, *args: Any, **kwargs: dict[str, Any]) -> None:
        """Run all callbacks for an event."""
        for listener in self._listeners.get(event_name, []):
            listener(*args, **kwargs)


class BaseSensorMixin:
    """Base sensor mixin."""

    _battery: float = 0
    _last_report: int | None = None
    _report_interval: int = 0

    @property
    def battery(self) -> Quantity[float]:
        """Return the battery in volts (V)."""
        return self._battery * UNIT_VOLTS

    @property
    def last_report(self) -> datetime | None:
        """Return last report."""
        return utc_timestamp_from_epoch(self._last_report)

    @property
    def report_interval(self) -> Quantity[int]:
        """Return the report interval in minutes."""
        return self._report_interval * UNIT_MINUTES


class AirSensorMixin(BaseSensorMixin):
    """Air sensor mixin."""

    _air_temperature: float = 0
    _last_lightning_strike_event: LightningStrikeEvent | None = None
    _lightning_strike_average_distance: float = 0
    _lightning_strike_count: int = 0
    _relative_humidity: float = 0
    _station_pressure: float = 0

    @property
    def air_temperature(self) -> Quantity[float]:
        """Return the air temperature in degrees Celsius (°C)."""
        return self._air_temperature * UNIT_DEGREES_CELSIUS

    @property
    def last_lightning_strike_event(self) -> LightningStrikeEvent | None:
        """Return the last lightning strike event."""
        return self._last_lightning_strike_event

    @property
    def lightning_strike_average_distance(self) -> Quantity[float]:
        """Return the lightning strike average distance in kilometers (km)."""
        return self._lightning_strike_average_distance * UNIT_KILOMETERS

    @property
    def lightning_strike_count(self) -> int:
        """Return the lightning strike count."""
        return self._lightning_strike_count

    @property
    def relative_humidity(self) -> Quantity[float]:
        """Return the relative humidity percentage."""
        return self._relative_humidity * UNIT_PERCENT

    @property
    def station_pressure(self) -> Quantity[float]:
        """Return the station pressure in millibars (mbar)."""
        return self._station_pressure * UNIT_MILLIBARS

    # Derived metrics

    @property
    def air_density(self) -> Quantity[float]:
        """Return the calculated air density in kilograms per cubic meter (kg/m³)."""
        return air_density(self.air_temperature, self.station_pressure)

    @property
    def delta_t(self) -> Quantity[float]:
        """Return the calculated delta t in delta degrees Celsius (Δ°C)."""
        return self.air_temperature - self.wet_bulb_temperature

    @property
    def dew_point_temperature(self) -> Quantity[float]:
        """Return the calculated dew point temperature in degrees Celsius (°C)."""
        return dew_point_temperature(self.air_temperature, self.relative_humidity)

    @property
    def heat_index(self) -> Quantity[float]:
        """Return the calculated heat index in degrees Celsius (°C)."""
        return heat_index(self.air_temperature, self.relative_humidity)

    @property
    def vapor_pressure(self) -> Quantity[float]:
        """Return the calculated vapor pressure in millibars (mbar)."""
        return vapor_pressure(self.air_temperature, self.relative_humidity)

    @property
    def wet_bulb_temperature(self) -> Quantity[float]:
        """Return the calculated wet bulb temperature in degrees Celsius (°C)."""
        return wet_bulb_temperature(
            self.air_temperature, self.relative_humidity, self.station_pressure
        )

    def calculate_sea_level_pressure(self, height: Quantity[float]) -> Quantity[float]:
        """Calculate the sea level pressure in millibars (mbar) from a specified height."""
        return sea_level_pressure(self.station_pressure, height, self.air_temperature)


class SkySensorMixin(BaseSensorMixin):
    """Sky sensor mixin."""

    _illuminance: int = 0
    _last_rain_start_event: RainStartEvent | None = None
    _last_wind_event: WindEvent | None = None
    _precipitation_type: int = 0
    _rain_accumulation_previous_minute: float = 0
    _solar_radiation: int = 0
    _uv: float = 0
    _wind_average: float = 0
    _wind_direction: int = 0
    _wind_gust: float = 0
    _wind_lull: float = 0
    _wind_sample_interval: int = 0

    @property
    def illuminance(self) -> Quantity[int]:
        """Return the illuminance is lux (lx)."""
        return self._illuminance * UNIT_LUX

    @property
    def last_rain_start_event(self) -> RainStartEvent | None:
        """Return the last rain start event."""
        return self._last_rain_start_event

    @property
    def last_wind_event(self) -> WindEvent | None:
        """Return the last wind event."""
        return self._last_wind_event

    @property
    def precipitation_type(self) -> PrecipitationType:
        """Return the precipitation type."""
        return PrecipitationType(self._precipitation_type)

    @property
    def rain_accumulation_previous_minute(self) -> Quantity[float]:
        """Return the rain accumulation from the previous minute in millimeters."""
        return self._rain_accumulation_previous_minute * UNIT_MILLIMETERS

    @property
    def rain_amount_previous_minute(self) -> Quantity[float]:
        """**Return the rain amount over previous minute in millimeters (mm/min).

        **This property has been deprecated. Please use `rain_accumulation_previous_minute`
        for an accumulation amount or `rain_rate` for an hourly intensity.
        """
        _LOGGER.warning(
            "The property 'rain_amount_previous_minute' has been deprecated due to "
            "inconsistent naming with the property and the units. Please use either "
            "`rain_accumulation_previous_minute` for a rain accumulation amount or "
            "`rain_rate` for an hourly intensity of rain."
        )
        return self._rain_accumulation_previous_minute * UNIT_MILLIMETERS_PER_MINUTE

    @property
    def rain_rate(self) -> Quantity[float]:
        """Return the rain rate in millimeters per hour (mm/hr).

        Based on the rain accumulation from the previous minute.
        """
        return self._rain_accumulation_previous_minute * 60 * UNIT_MILLIMETERS_PER_HOUR

    @property
    def solar_radiation(self) -> Quantity[int]:
        """Return the solar radiation in watts per square meter (W/m²)."""
        return self._solar_radiation * UNIT_IRRADIATION

    @property
    def uv(self) -> float:  # pylint: disable=invalid-name
        """Return the uv index."""
        return self._uv

    @property
    def wind_average(self) -> Quantity[float]:
        """Return the wind average in meters per second (m/s)."""
        return self._wind_average * UNIT_METERS_PER_SECOND

    @property
    def wind_direction(self) -> Quantity[int]:
        """Return the wind direction in degrees."""
        return self._wind_direction * UNIT_DEGREES

    @property
    def wind_direction_cardinal(self) -> str:
        """Return the wind direction cardinality."""
        return degrees_to_cardinal(self._wind_direction)

    @property
    def wind_gust(self) -> Quantity[float]:
        """Return the wind gust in meters per second (m/s)."""
        return self._wind_gust * UNIT_METERS_PER_SECOND

    @property
    def wind_lull(self) -> Quantity[float]:
        """Return the wind lull in meters per second (m/s)."""
        return self._wind_lull * UNIT_METERS_PER_SECOND

    @property
    def wind_sample_interval(self) -> Quantity[int]:
        """Return wind sample interval in seconds."""
        return self._wind_sample_interval * UNIT_MINUTES

    @property
    def wind_speed(self) -> Quantity[float]:
        """Return the most recent wind speed in meters per second (m/s)."""
        return (
            self.wind_average
            if self._last_wind_event is None
            or (self._last_report or 0) > self._last_wind_event.epoch
            else self._last_wind_event.speed
        )



"""Enums module."""
from __future__ import annotations

from enum import IntEnum, unique


@unique
class PrecipitationType(IntEnum):
    """Precipitation types."""

    NONE = 0
    RAIN = 1
    HAIL = 2
    RAIN_HAIL = 3

    # Handle unknown/future fan modes
    UNKNOWN = -1

    @classmethod
    def _missing_(cls, _: object) -> PrecipitationType:  # pragma: no cover
        """Return default if not found."""
        return cls.UNKNOWN


@unique
class RadioStatus(IntEnum):
    """Radio statuses."""

    RADIO_OFF = 0
    RADIO_ON = 1
    RADIO_ACTIVE = 3
    BLE_CONNECTED = 7

    # Handle unknown/future fan modes
    UNKNOWN = -1

    @classmethod
    def _missing_(cls, _: object) -> RadioStatus:  # pragma: no cover
        """Return a default if not found."""
        return cls.UNKNOWN
